import { getPassword, setPassword } from 'keytar';
import { parse as parseCookies, serialize as serializeCookie } from 'cookie';
import * as inquirer from 'inquirer';
const SERVICE = 'HttpProxySecureCookies';
/**
 * Serialize cookies object to the cookie string as seen in request headers.
 */
function toCookieString(cookies) {
    return Object.entries(cookies)
        .map(([key, value]) => serializeCookie(key, value))
        .join('; ');
}
/**
 * Parse the `Set-Cookie` header from remote servers.
 */
function parseSetCookie(cookieHeaders) {
    if (!cookieHeaders)
        return [[], {}];
    const cookieArray = Array.isArray(cookieHeaders)
        ? cookieHeaders
        : [String(cookieHeaders)];
    const cookies = {};
    return [
        cookieArray,
        cookieArray
            .map((x) => x.split(';', 1)[0].split('=', 2))
            .reduce((acc, [key, value]) => {
            acc[key] = value;
            return acc;
        }, cookies),
    ];
}
/**
 * Generate a proxy config that points to a remote host.
 */
export default function secureCookieProxy(options) {
    const { target, keychainAccount, unauthorizedStatusCode = [401], cookieRewrite, ...restOptions } = typeof options === 'string' ? { target: options } : options;
    // default account is the target sans protocol
    const account = keychainAccount || target.split('://')[1];
    const unauthroizedCode = new Set(Array.isArray(unauthorizedStatusCode)
        ? unauthorizedStatusCode
        : [unauthorizedStatusCode]);
    let secureCookies;
    let waitingForInput = false;
    async function getFromKeyChain() {
        return getPassword(SERVICE, account);
    }
    async function askForUserInput(message) {
        waitingForInput = true;
        const { cookieString } = await inquirer.prompt([
            {
                type: 'password',
                name: 'cookieString',
                message,
            },
        ]);
        await setPassword(SERVICE, account, cookieString);
        waitingForInput = false;
        console.log('\nSuccessfully saved your cookie. Please refresh.');
        return cookieString;
    }
    /**
     * Read cookies from keychain or user input.
     */
    async function getCookies(message) {
        if (waitingForInput)
            return null;
        try {
            secureCookies = parseCookies((message ? await askForUserInput(message) : await getFromKeyChain()) ||
                '');
        }
        catch {
            await getCookies('Invalid cookie string, please try again: ');
        }
        return secureCookies;
    }
    /**
     * Add secure cookies to proxy request header
     */
    function addCookie(proxyReq, req) {
        proxyReq.setHeader('cookie', toCookieString({
            ...secureCookies,
            ...parseCookies(req.headers.cookie || ''),
        }));
    }
    // quitely get initial cookies
    getCookies();
    return {
        secure: false,
        changeOrigin: true,
        target,
        ws: target.startsWith('ws'),
        ...restOptions,
        onProxyReq: addCookie,
        onProxyReqWs: addCookie,
        onProxyRes(proxyRes, req, res) {
            // update cookies if API returns 401
            if (proxyRes.statusCode && unauthroizedCode.has(proxyRes.statusCode)) {
                getCookies(`
Authentication failed for ${target}${req.url}

You either haven't provide an auth cookie or it expired.
Please login to ${target} and copy the HTTP cookie string here.

It will be securely stored in system keychain:`);
            }
            // add missing cookies to the response so they can be used on the client
            // side as well.
            if (secureCookies) {
                const reqCookies = parseCookies(req.headers.cookie || '');
                const [cookieHeaders, resCookies] = parseSetCookie(res.getHeader('set-cookie'));
                const clientCookies = cookieRewrite
                    ? cookieRewrite(secureCookies)
                    : secureCookies;
                let hasMissingCookie = false;
                Object.keys(clientCookies).forEach((key) => {
                    if (!(key in reqCookies) && !(key in resCookies)) {
                        // set all secure cookie to the root path
                        cookieHeaders.push(serializeCookie(key, clientCookies[key], { path: '/' }));
                        hasMissingCookie = true;
                    }
                });
                if (hasMissingCookie) {
                    res.setHeader('Set-Cookie', cookieHeaders);
                }
            }
        },
    };
}
